import json
import ipaddress as ip
from typing import Iterable, Optional, Sequence, SupportsBytes, SupportsIndex

import pyarrow as pa


class IPScalar(pa.ExtensionScalar):
    def as_py(self) -> ip.IPv4Address | ip.IPv6Address | None:
        return None if self.value is None else unpack_ip(self.value.as_py())


class IPType(pa.ExtensionType):
    # NOTE: The identifier for the extension type of VAST's ip type has not
    # changed when the type was renamed from address to ip because that would be
    # a breaking change. This is fixable by registering two separate extension
    # types with the same functionality but different ids, but that's a lot of
    # effort for something users don't usually see.
    ext_name = "vast.address"
    ext_type = pa.binary(16)

    def __init__(self):
        pa.ExtensionType.__init__(self, self.ext_type, self.ext_name)

    def __arrow_ext_serialize__(self) -> bytes:
        return self.ext_name.encode()

    @classmethod
    def __arrow_ext_deserialize__(self, storage_type, serialized: bytes):
        if serialized.decode() != self.ext_name:
            raise TypeError("type identifier does not match")
        if storage_type != self.ext_type:
            raise TypeError("storage type does not match")
        return IPType()

    def __reduce__(self):
        return IPScalar, ()

    def __arrow_ext_scalar_class__(self):
        return IPScalar


class SubnetScalar(pa.ExtensionScalar):
    def as_py(self) -> ip.IPv4Network | ip.IPv6Network | None:
        address = self.value[0].as_py()
        length = self.value[1].as_py()
        if address is None or length is None:
            return None
        return ip.ip_network((address, length), strict=False)


class SubnetType(pa.ExtensionType):
    ext_name = "vast.subnet"
    ext_type = pa.struct([("address", IPType()), ("length", pa.uint8())])

    def __init__(self):
        pa.ExtensionType.__init__(self, self.ext_type, self.ext_name)

    def __arrow_ext_serialize__(self) -> bytes:
        return self.ext_name.encode()

    @classmethod
    def __arrow_ext_deserialize__(self, storage_type, serialized: bytes):
        if serialized.decode() != self.ext_name:
            raise TypeError("type identifier does not match")
        if storage_type != self.ext_type:
            raise TypeError("storage type does not match")
        return SubnetType()

    def __reduce__(self):
        return SubnetScalar, ()

    def __arrow_ext_scalar_class__(self):
        return SubnetScalar


class EnumScalar(pa.ExtensionScalar):
    def as_py(self) -> str | None:
        return None if self.value is None else self.value.as_py()


class EnumType(pa.ExtensionType):
    """An extension class for enum data types.

    `EnumType` stores as name-to-value mapping called `fields` that maps to the
    underlying integer type of represented the enum."""

    # VAST's flatbuffer type representation uses a 32-bit unsigned integer. We
    # use an 8-bit type here only for backwards compatibility to the legacy
    # type. Eventually this will be a 32-bit type as well.
    DICTIONARY_INDEX_TYPE = pa.uint8()

    ext_name = "vast.enumeration"
    ext_type = pa.dictionary(DICTIONARY_INDEX_TYPE, pa.string())

    def __init__(self, fields: dict[str, int]):
        self._fields = fields
        pa.ExtensionType.__init__(self, self.ext_type, self.ext_name)

    @property
    def fields(self) -> dict[str, int]:
        return self._fields

    def __arrow_ext_serialize__(self) -> bytes:
        return json.dumps(self._fields).encode()

    @classmethod
    def __arrow_ext_deserialize__(self, storage_type, serialized: bytes):
        fields = json.loads(serialized.decode())
        if storage_type != self.ext_type:
            raise TypeError("storage type does not match")
        return EnumType(fields)

    def __reduce__(self):
        return EnumScalar, ()

    def __arrow_ext_scalar_class__(self):
        return EnumScalar


def names(schema: pa.Schema):
    meta = schema.metadata
    return [meta[key].decode() for key in meta if key.startswith(b"VAST:name:")]


def name(schema: pa.Schema):
    xs = names(schema)
    return xs[0] if xs[0] else ""


def pack_ip(address: str | ip.IPv4Address | ip.IPv6Address) -> bytes:
    match address:
        case str():
            return pack_ip(ip.ip_address(address))
        case ip.IPv4Address():
            prefix = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff"
            return prefix + address.packed
        case ip.IPv6Address():
            return address.packed


# Accepts a 128-bit buffer holding an IPv6 address and returns an IPv4 or IPv6
# address.
def unpack_ip(
    buffer: Iterable[SupportsIndex] | SupportsBytes,
) -> ip.IPv4Address | ip.IPv6Address:
    num = int.from_bytes(buffer, byteorder="big")
    # Convert IPv4 mapped addresses back to regular IPv4.
    # https://tools.ietf.org/html/rfc4291#section-2.5.5.2
    if (num >> 32) == 65535:
        num = num - (65535 << 32)
    return ip.ip_address(num)


def pack_subnet(
    subnet: str | ip.IPv4Network | ip.IPv6Network | None,
) -> tuple[bytes | None, int | None]:
    if subnet is None:
        return (None, None)
    match subnet:
        case str():
            return pack_subnet(ip.ip_network(subnet))
        case ip.IPv4Network():
            prefix = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff"
            return (prefix + subnet.network_address.packed, subnet.prefixlen)
        case ip.IPv6Network():
            return (subnet.network_address.packed, subnet.prefixlen)


VastExtensionType = IPType | EnumType | SubnetType


def py_dict_to_arrow_dict(dict: dict[str, int]) -> pa.StringArray:
    """Build an Arrow dictionary array from a Python dictionary

    - Values should be greater or equal to 0.
    - Non-consecutive values will interpolated with null."""
    dictionary_py: list[Optional[str]] = [None] * (max(dict.values()) + 1)
    for (value, idx) in dict.items():
        assert idx >= 0, f"Dictionary indices should be >=0, got {idx}"
        dictionary_py[idx] = value
    return pa.array(dictionary_py, pa.string())


def extension_array(obj: Sequence, type: pa.DataType) -> pa.Array:
    """Create a `pyarrow.Array` from a Python object, using an
    `pyarrow.ExtensionArray` if the type is an extension type."""
    match type:
        case IPType():
            arr = [pack_ip(e) for e in obj]
            storage = pa.array(arr, pa.binary(16))
            return pa.ExtensionArray.from_storage(type, storage)
        case SubnetType():
            arrs = list(zip(*[pack_subnet(e) for e in obj]))
            addr_storage = pa.array(arrs[0], pa.binary(16))
            addr_array = pa.ExtensionArray.from_storage(IPType(), addr_storage)
            prefix_array = pa.array(arrs[1], pa.uint8())
            storage = pa.StructArray.from_arrays(
                [
                    addr_array,
                    prefix_array,
                ],
                names=["address", "length"],
            )
            return pa.ExtensionArray.from_storage(type, storage)
        case EnumType(fields=fields):
            # use the mappings in the `fields` metadata as indices for building
            # the dictionary
            indices_py = [None if e is None else fields[e] for e in obj]
            indices = pa.array(indices_py, EnumType.DICTIONARY_INDEX_TYPE)
            dictionary = py_dict_to_arrow_dict(fields)
            storage = pa.DictionaryArray.from_arrays(indices, dictionary)
            return pa.ExtensionArray.from_storage(type, storage)
        case _:
            return pa.array(obj, type)


# Modules are intialized exactly once, so we can perform the registration here.
pa.register_extension_type(IPType())
pa.register_extension_type(SubnetType())
pa.register_extension_type(EnumType({"stub": 0}))



import asyncio
from datetime import datetime
import hashlib
import json
import os
import time
from typing import Optional

import aiohttp
import pyvast.utils.logging as logging

from pyvast import VAST, ExportMode, to_json_rows

logger = logging.get("vast.thehive.app")

THEHIVE_ORGADMIN_EMAIL = os.environ["DEFAULT_ORGADMIN_EMAIL"]
THEHIVE_ORGADMIN_PWD = os.environ["DEFAULT_ORGADMIN_PWD"]
THEHIVE_URL = os.environ["THEHIVE_URL"]
BACKFILL_LIMIT = int(os.environ["BACKFILL_LIMIT"])


async def call_thehive(
    path: str,
    payload: Optional[dict] = None,
) -> str:
    """Call a TheHive endpoint with basic auth"""
    path = f"{THEHIVE_URL}{path}"
    auth = aiohttp.BasicAuth(THEHIVE_ORGADMIN_EMAIL, THEHIVE_ORGADMIN_PWD)
    async with aiohttp.ClientSession() as session:
        if payload is None:
            resp = await session.get(path, auth=auth)
        else:
            resp = await session.post(path, json=payload, auth=auth)

    resp_txt = await resp.text()
    logger.debug(f"Resp to query on TheHive API {path}: {resp_txt}")
    resp.raise_for_status()
    logger.info(f"Call to TheHive API {path} successful!")
    return resp_txt


async def wait_for_thehive(
    path: str,
    timeout: int,
    payload: Optional[dict] = None,
) -> str:
    """Call thehive repeatedly until timeout"""
    start = time.time()
    while True:
        try:
            return await call_thehive(path, payload)
        except Exception as e:
            if time.time() - start > timeout:
                raise Exception("Timed out trying to reach TheHive")
            logger.debug(e)
            await asyncio.sleep(1)


def suricata2thehive(event: dict) -> dict:
    """Convert a Suricata alert event into a TheHive alert"""
    sighted_time_iso = event.get("timestamp", datetime.now().isoformat())
    start_time_iso = event.get("flow", {}).get("timestamp", sighted_time_iso)
    alert = event.get("alert", {})
    # Severity is defined differently:
    # - in Suricata: 1-255 (usually 1-4), 1 being the highest
    # - in Thehive: 1-4, 4 being the highest
    severity = max(5 - alert.get("severity", 3), 1)
    category = alert.get("category")
    desc = f'{alert.get("signature_id", "No signature ID")}: {alert.get("signature", "No signature")}'
    # A unique identifier of this alert, hashing together the start time and flow id
    src_ref = hashlib.md5(
        f'{start_time_iso}{event.get("flow_id", "")}'.encode()
    ).hexdigest()

    return {
        "type": category,
        "source": "suricata",
        "sourceRef": src_ref,
        "title": "Suricata Alert",
        "description": desc,
        "severity": severity,
        "date": sighted_time_iso,
        "tags": [],
        "observables": [
            {
                "dataType": "ip",
                "data": event["src_ip"],
                "message": "Source IP",
                "startDate": start_time_iso,
                "tags": [],
                "ioc": False,
                "sighted": True,
                "sightedAt": sighted_time_iso,
                "ignoreSimilarity": False,
            },
            {
                "dataType": "ip",
                "data": event["dest_ip"],
                "message": "Destination IP",
                "startDate": start_time_iso,
                "tags": [],
                "ioc": False,
                "sighted": True,
                "sightedAt": sighted_time_iso,
                "ignoreSimilarity": False,
            },
        ],
    }


async def on_suricata_alert(alert: dict):
    global SENT_ALERT_REFS
    logger.debug(f"Received Suricata alert: {alert}")
    thehive_alert = suricata2thehive(alert)
    logger.debug(f"Resulting TheHive alert: {thehive_alert}")

    ref = thehive_alert["sourceRef"]
    try:
        await call_thehive("/api/v1/alert", thehive_alert)
        logger.debug(f"Alert with hash {ref} ingested")
    except aiohttp.ClientResponseError as e:
        if e.code == 400:
            logger.debug(f"Alert with hash {ref} not ingested (error 400)")
        else:
            raise e


async def run_async():
    vast_cli = VAST()
    await vast_cli.status(60, retry_delay=1)
    await wait_for_thehive("/api/v1/user/current", 180)
    expr = '#type == "suricata.alert"'
    # We don't use "UNIFIED" to specify a limit on the HISTORICAL backfill
    logger.info("Starting retro filling...")
    hist_iter = vast_cli.export(expr, ExportMode.HISTORICAL, limit=BACKFILL_LIMIT)
    async for row in to_json_rows(hist_iter):
        await on_suricata_alert(row.data)
    logger.info("Starting live forwarding...")
    cont_iter = vast_cli.export(expr, ExportMode.CONTINUOUS)
    async for row in to_json_rows(cont_iter):
        await on_suricata_alert(row.data)


def run():
    logger.info("Starting TheHive app...")
    asyncio.run(run_async())
    logger.info("TheHive app stopped")


async def wait_for_alerts(timeout):
    """Call TheHive listAlert API until the number of alerts is greater than 0"""
    list_query = {"query": [{"_name": "listAlert"}]}
    start = time.time()
    while True:
        try:
            alerts_json = await call_thehive("/api/v1/query", list_query)
            alert_count = len(json.loads(alerts_json))
            if alert_count == 0:
                raise Exception("No alerts in TheHive yet")
            return alert_count
        except Exception as e:
            if time.time() - start > timeout:
                raise Exception("Timed out trying to reach TheHive")
            sleep_duration = 2
            logger.debug(f"{e} Retryinng in {sleep_duration} second(s)")
            await asyncio.sleep(sleep_duration)


def count_alerts():
    """Wait for alerts in TheHive then print their count"""
    nb_alerts = asyncio.run(wait_for_alerts(180))
    logger.info(f"alert_count={nb_alerts}")



import base64
from collections import defaultdict
from dataclasses import dataclass
import datetime
from decimal import Decimal
from typing import Any, AsyncIterable

import numpy as np
import pyarrow as pa
import pyvast.utils.arrow
import pyvast.utils.logging

from .vast import TableSlice

logger = pyvast.utils.logging.get("vast.vast")

_JSON_COMPATIBILITY_DOCSTRING_ = """JSON types are numbers, booleans, strings, arrays and objects
- dates and times are formated with ISO 8601
- raw bytes are Base64 encoded"""


def to_pyarrow(batch: TableSlice) -> pa.RecordBatch:
    """Represent the provided TableSlice as a PyArrow RecordBatch"""
    from .vast import PyArrowTableSlice

    if not isinstance(batch, PyArrowTableSlice):
        raise TypeError(f"Cannot convert {type(batch)} to pyarrow.RecordBatch")
    return batch._batch


async def collect_pyarrow(
    stream: AsyncIterable[TableSlice],
) -> dict[str, list[pa.Table]]:
    """Iterate through the TableSlice stream and sort the record batches by
    schema as lists of PyArrow Tables with identical schemas"""
    num_batches = 0
    num_rows = 0
    batches = defaultdict(list)
    async for slice in stream:
        batch = to_pyarrow(slice)
        name = pyvast.utils.arrow.name(batch.schema)
        logger.debug(f"got batch of {name}")
        num_batches += 1
        num_rows += batch.num_rows
        # TODO: this might be slow, but schemas are not
        # hashable. The string representation is the next best
        # thing to determine uniqueness.
        batches[batch.schema.to_string()].append(batch)
    result = defaultdict(list)
    for (_, batches) in batches.items():
        table = pa.Table.from_batches(batches)
        name = pyvast.utils.arrow.name(table.schema)
        result[name].append(table)
    return result


@dataclass
class VastRow:
    """A row wise representation of the data and metadata

    - name: the VAST type for the row
    - data: the event data contained in the row"""

    name: str
    data: dict[str, Any]


def arrow_dict_to_json_dict(dictionary):
    f"""Convert a result item of PyArrow to_pylist into a dictionary with
    JSON compatible value types

    {_JSON_COMPATIBILITY_DOCSTRING_}"""
    for key, value in dictionary.items():
        if isinstance(value, dict):
            arrow_dict_to_json_dict(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, (dict, list)):
                    arrow_dict_to_json_dict(item)
                elif not isinstance(item, (int, float, bool)):
                    item = str(item)
        elif isinstance(value, (np.floating, Decimal)):
            dictionary[key] = float(value)
        elif isinstance(value, bytes):
            dictionary[key] = base64.b64encode(value).decode()
        elif isinstance(value, (datetime.time, datetime.datetime, datetime.date)):
            dictionary[key] = value.isoformat()
        elif value is None or isinstance(value, (int, float, bool)):
            dictionary[key] = value
        elif not isinstance(value, (int, float, bool)):
            dictionary[key] = str(value)
    return dictionary


async def to_json_rows(
    stream: AsyncIterable[TableSlice],
) -> AsyncIterable[VastRow]:
    f"""Convert the TableSlice iterator to a row by row iterator with value types
    dumbed down to JSON compatible types

    {_JSON_COMPATIBILITY_DOCSTRING_}"""
    async for slice in stream:
        batch = to_pyarrow(slice)
        name = pyvast.utils.arrow.name(batch.schema)
        for row in batch.to_pylist():
            yield VastRow(name, arrow_dict_to_json_dict(row))



from dynaconf import Dynaconf, Validator

# TODO: make the path configurable
CONFIG_FILES = ["config.yaml", "config.yml"]
PREFIX = "python."


class Config:
    def __init__(self, conf: Dynaconf):
        self.conf = conf

    def get(self, key):
        return self.conf.get(PREFIX + key)


def create() -> Config:
    settings = Dynaconf(
        settings_files=CONFIG_FILES,
        load_dotenv=True,
        envvar_prefix="VAST",
    )
    settings.validators.register(
        Validator(PREFIX + "console-verbosity", default="debug"),
        Validator(PREFIX + "file-verbosity", default="quiet"),
        Validator(PREFIX + "log-file", default="vast.log"),
    )
    settings.validators.validate_all()
    return Config(settings)
